#define NOMINMAX
#include <fstream>
#include <DirectXMath.h>
#include <memory.h>
#include <math.h>

using namespace DirectX;

struct CorridorMapNode
{
    XMFLOAT2 Position;
    float    Radius;
};

float InitializeSolverConstraints(
    XMVECTOR* c, float* r,
    const CorridorMapNode* subgraph, int subgraphN,
    FXMVECTOR startPosition, FXMVECTOR startTangent,
    FXMVECTOR goalPosition, GXMVECTOR goalTangent)
{
    // Determine the discretization resolution h as the average waypoint separation distance.
    XMVECTOR initialPathLength = XMVectorZero();
    for (int i = 1; i < subgraphN; i++)
        initialPathLength += XMVector2Length(XMLoadFloat2(&(subgraph[i].Position)) - XMLoadFloat2(&(subgraph[i - 1].Position)));

    float h = XMVectorGetX(initialPathLength) / (float)(subgraphN - 1);

    // c and r are arrays of length subgraphN + 4.
    c[0] = startPosition - startTangent * h;
    r[0] = 0;
    c[1] = startPosition;
    r[1] = 0;

    for (int i = 0; i < subgraphN; i++)
    {
        c[i + 2] = XMLoadFloat2(&(subgraph[i].Position));
        r[i + 2] = subgraph[i].Radius;
    }

    c[subgraphN + 2] = goalPosition;
    r[subgraphN + 2] = 0;
    c[subgraphN + 3] = goalPosition + goalTangent * h;
    r[subgraphN + 3] = 0;

    return h;
}

void InitializeSolverWeights(
    float* w, int n,
    FXMVECTOR startTangent, FXMVECTOR goalTangent,
    float ws, float wm, float we)
{
    // A zero length start/goal tangent vector means the agent facing direction is
    // arbitrary there and should be ignored.
    bool ignoreStartTangent = XMVectorGetX(XMVector2LengthSq(startTangent)) == 0;
    bool ignoreGoalTangent  = XMVectorGetX(XMVector2LengthSq(goalTangent)) == 0;

    // Use the boosted weights ws/we at start/goal waypoints to align better with the path
    // tangent there, unless they are zero length vectors.
    if (ignoreStartTangent)
        ws = wm;
    if (ignoreGoalTangent)
        we = wm;

    // The first and last waypoint are dummy waypoints for the agent facing constraint, which
    // have a corresponding zero weight.
    w[0] = 0;
    for (int i = 1; i < n - 1; i++)
    {
        // The weight is ws/we at the start/goal waypoints, and has a 4th power falloff to
        // wm at the middle waypoint.
        float wt = powf(fabsf(((float)(i - 1) / (float)(n - 3)) * 2 - 1), 4);
        if ((float)(i - 1) / (float)(n - 3) <= 0.5f)
            w[i] = ws * wt + wm * (1 - wt);
        else
            w[i] = we * wt + wm * (1 - wt);
    }
    w[n - 1] = 0;

    // Zero out the weights at the start/goal point if the tangents are not specified. This has the effect that
    // no smoothing is done there in that case.
    if (ignoreStartTangent)
        w[1] = 0;
    if (ignoreGoalTangent)
        w[n - 2] = 0;
}

float InitializeSolverStepsizes(
    float* __restrict sigma, float* __restrict tau, int n,
    float alpha, float beta, const float* __restrict w, float h)
{
    const float twoPowAlpha = powf(2.0f, 2 - alpha);
    const float hPowAlpha   = powf(h,    2 - alpha);

    // The first and last waypoint are dummy waypoints for the agent facing constraint, which
    // have a corresponding zero step size.
    sigma[0] = 0;
    tau[0]   = 0;

    float wPowAlphaPrev = 0;
    float wPowAlphaCurr = powf(w[1], 2 - alpha);
    for (int i = 1; i < n - 1; i++)
    {
        if (w[i] != 0)
            sigma[i] = (twoPowAlpha * (h * h) * wPowAlphaCurr) / (2 * hPowAlpha * (w[i] * w[i]) * beta * (twoPowAlpha + 2));
        else
            sigma[i] = 0;

        float wPowAlphaNext = powf(w[i + 1], 2 - alpha);
        tau[i] = (hPowAlpha * beta) / (2 + wPowAlphaPrev + twoPowAlpha * wPowAlphaCurr + wPowAlphaNext);

        wPowAlphaPrev = wPowAlphaCurr;
        wPowAlphaCurr = wPowAlphaNext;
    }

    sigma[n - 1] = 0;
    tau[n - 1]   = 0;

    return (h * h) / (2 * hPowAlpha * beta);
}

void SolverIterations(
    int k, int n,
    XMVECTOR* __restrict v, XMVECTOR* __restrict vhat, XMVECTOR* __restrict p, XMVECTOR* __restrict q,
    const XMVECTOR* __restrict c, const float* __restrict r, const float* __restrict w,
    const float* __restrict sigma, float mu, const float* __restrict tau, float h)
{
    // On the first iteration of the minimizer, v = c, vhat = c, p = 0 and q = 0.
    memcpy(v,    c, n * sizeof(XMVECTOR));
    memcpy(vhat, c, n * sizeof(XMVECTOR));
    memset(p,    0, n * sizeof(XMVECTOR));
    memset(q,    0, n * sizeof(XMVECTOR));

    do
    {
        const float rcph = 1 / h;

        q[0] += mu * rcph * (vhat[1] - vhat[0]);
        XMVECTOR qL2sqr = XMVector2LengthSq(q[0]);
        for (int i = 1; i < n - 1; i++)
        {
            XMVECTOR bckdiffvhat = rcph * (vhat[i] - vhat[i - 1]);
            XMVECTOR fwddiffvhat = rcph * (vhat[i + 1] - vhat[i]);

            // Gradient ascent on the dual variable p -> p + sigma * K1.vhat
            p[i] += sigma[i] * w[i] * (bckdiffvhat - fwddiffvhat);
            // Proximal map of the dual variable p.
            p[i] /= XMVectorSplatOne() + XMVectorReplicate(sigma[i]);

            // Gradient ascent on the dual variable q -> q + mu * K2.vhat
            q[i] += mu * fwddiffvhat;
            qL2sqr += XMVector2LengthSq(q[i]);
        }

        // Proximal map of the dual variable q.
        if (XMVectorGetX(qL2sqr) > 1)
        {
            XMVECTOR normalizeq = XMVectorReciprocal(XMVectorSqrt(qL2sqr));
            for (int i = 0; i < n - 1; i++)
                q[i] *= normalizeq;
        }

        for (int i = 1; i < n - 1; i++)
        {
            XMVECTOR sumdiffpw = 2 * p[i] * w[i] - p[i - 1] * w[i - 1] - p[i + 1] * w[i + 1];
            XMVECTOR bckdiffq = q[i] - q[i - 1];

            // Gradient descent on the primal variable v -> v - tau * Kstar.(p,q)
            XMVECTOR vnext = v[i] - tau[i] * rcph * (sumdiffpw - bckdiffq);
            // Proximal map of the primal variable v.
            XMVECTOR dvc = vnext - c[i];
            // Add FLT_EPSILON to radius in denominator avoid 0/0 case when the radius and length of dvc are both zero.
            vnext = c[i] + dvc * r[i] / XMVectorMax(XMVectorReplicate(r[i] + FLT_EPSILON), XMVector2Length(dvc));

            // Approximate extragradient.
            vhat[i] = 2 * vnext - v[i];
            v[i] = vnext;
        }

        k--;
    }
    while (k > 0);
}

void DumpPathAsSVG(const char* filename, int n, const XMVECTOR* v, const XMVECTOR* c, const float* r)
{
    std::ofstream s(filename);
    s << std::fixed;
    s << "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>" << std::endl;
    s << "<svg xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\">" << std::endl;

    for (int i = 2; i < n - 2; i++)
        s << "<circle cx=\"" << XMVectorGetX(c[i]) << "\" cy=\"" << XMVectorGetY(c[i]) << "\" r=\"" << r[i] << "\" fill=\"none\" stroke=\"gray\"/>" << std::endl;

    s << "<polyline points=\"";
    for (int i = 1; i < n - 1; i++)
        s << XMVectorGetX(v[i]) << ',' << XMVectorGetY(v[i]) << " ";
    s << "\" stroke=\"red\" fill=\"none\"/>";

    s << "</svg>";
}

void main()
{
    static const CorridorMapNode subgraph[] = {
        { { 65.f,44.f },31.f },{ { 79.f,43.f },29.4279f },{ { 93.f,42.f },28.0713f },{ { 110.f,40.f },27.f },
        { { 124.f,39.f },25.4951f },{ { 138.f,38.f },24.1868f },{ { 152.f,37.f },22.8035f },{ { 167.f,37.f },23.7697f },
        { { 188.f,46.f },31.9061f },{ { 184.f,66.f },24.1868f },{ { 177.f,84.f },22.0227f },{ { 165.f,96.f },26.4764f },
        { { 147.f,109.f },36.1248f },{ { 125.f,134.f },56.8859f },{ { 101.f,170.f },88.f },{ { 110.f,189.f },97.f },
        { { 119.f,208.f },106.f },{ { 131.f,229.f },117.618f },{ { 172.f,211.f },85.3815f },{ { 198.f,194.f },66.7308f },
        { { 214.f,179.f },56.8595f },{ { 228.f,161.f },50.2494f },{ { 237.f,149.f },50.9117f },{ { 249.f,136.f },54.5619f },
        { { 264.f,124.f },61.6604f },{ { 285.f,112.f },73.6817f },{ { 312.f,103.f },89.8109f },{ { 328.f,107.f },92.6607f },
        { { 345.f,112.f },98.0816f },{ { 363.f,118.f },105.f },{ { 368.f,136.f },99.6444f },{ { 341.f,182.f },58.3095f },
        { { 316.f,205.f },36.3456f },{ { 301.f,221.f },28.1603f },{ { 295.f,234.f },26.9258f },{ { 291.f,253.f },29.7321f },
        { { 290.f,272.f },35.1141f },{ { 297.f,298.f },50.f },{ { 326.f,284.f },30.1496f },{ { 349.f,268.f },15.f },
        { { 382.f,279.f },25.0799f },{ { 408.f,301.f },47.f },{ { 421.f,300.f },47.f }
    };

    static const XMVECTORF32 startPosition = { 58.f, 67.f };
    static const XMVECTORF32 startFacingDirection = { -0.316228f, -0.948683f };
    static const XMVECTORF32 goalPosition = { 420.f, 280.f };
    static const XMVECTORF32 goalFacingDirection = { 0.707107f, -0.707107f };

    const float startSmoothingWeight = 10;
    const float goalSmoothingWeight = 10;
    const float smoothingWeight = 2;
    const float stepSizeAlpha = 0.0625f;
    const float stepSizeBeta = 10;
    const int solverIterationCount = 100;

    const int subgraphN = sizeof(subgraph) / sizeof(subgraph[0]);

    // The corridor map subgraph is prepended and appended with the start and goal positions to
    // construct the initial (non-optimal) path. This initial path is extended at both ends with a
    // dummy additional waypoint for the agent facing constraint at the start and goal. Hence, the
    // solver arrays are of length subgraphN + 4.
    const int n = subgraphN + 4;

    // The solver state is allocated as a single block of aligned memory.
    void* alignedBlock = _aligned_malloc(n * 5 * sizeof(XMVECTOR) + n * 4 * sizeof(float), __alignof(XMVECTOR));
    XMVECTOR* __restrict c = (XMVECTOR*)alignedBlock;
    XMVECTOR* __restrict v = c + n;
    XMVECTOR* __restrict vhat = v + n;
    XMVECTOR* __restrict p = vhat + n;
    XMVECTOR* __restrict q = p + n;
    float*    __restrict r = (float*)(q + n);
    float*    __restrict w = r + n;
    float*    __restrict sigma = w + n;
    float*    __restrict tau = sigma + n;

    float h = InitializeSolverConstraints(c, r, subgraph, subgraphN, startPosition, startFacingDirection, goalPosition, goalFacingDirection);
    InitializeSolverWeights(w, n, startFacingDirection, goalFacingDirection, startSmoothingWeight, smoothingWeight, goalSmoothingWeight);
    float mu = InitializeSolverStepsizes(sigma, tau, n, stepSizeAlpha, stepSizeBeta, w, h);

    DumpPathAsSVG("initial.svg", n, c, c, r);
    SolverIterations(solverIterationCount, n, v, vhat, p, q, c, r, w, sigma, mu, tau, h);
    DumpPathAsSVG("smoothed.svg", n, v, c, r);

    _aligned_free(alignedBlock);
}
